from loguru import logger
import psycopg2
import psycopg2.extras
import os


MAX_RETRY_CNT = 3

class PgDbHandleException(Exception):
    def __init__(self, errType, code, msg):
        super().__init__()
        self.errType = errType
        self.code = code
        self.msg = msg

class PgDbHandle(object):
    def __init__(self, config):
        logger.info("constructor called")
        self.config = config
        self.connection = None
    
    def connect(self):
        dbHost = self.config["database"]["host"]
        dbUser = self.config["database"]["user"]
        dbPassword = self.config["database"]["password"]
        if dbPassword == "$ENV$":
            dbPassword = os.environ["DB_PASSWORD"]
        dbName = self.config["database"]["name"]

        logger.info(f"Connect to database")
        self.connInfo = f"user={dbUser} host={dbHost} dbname={dbName} sslmode=require"
        if dbPassword != "$NONE$":
            self.connInfo += f" password={dbPassword}"
        self.connection = psycopg2.connect(self.connInfo)
        self.connection.autocommit = False
        logger.info("Connection to database established")

    def disconnect(self):
        logger.info(f"Disonnect from database")
        if self.connection:
            self.connection.close()

    def execute(self, func, *args):
        try:
            retryCnt = 0
            while True:
                try:
                    if self.connection.status != psycopg2.extensions.STATUS_READY:
                        logger.error("resetting database connection")
                        self.connection.reset()
                    if self.connection.closed != 0:
                        logger.error("re-opening database connection")
                        self.connection = psycopg2.connect(self.connInfo)
                        self.connection.autocommit = False
                    with self.connection as conn:
                        with conn.cursor() as cursor:
                            func(cursor, *args)
                    break # in case of success we leave this retry loop
                except psycopg2.InterfaceError as e:
                    retryCnt += 1
                    logger.error(f"InterfaceError when talking to database: {e}, {retryCnt}. retry of {MAX_RETRY_CNT}")
                    if retryCnt > MAX_RETRY_CNT:
                        logger.error("Giving up")
                        raise e
        except psycopg2.Error as e:
            raise PgDbHandleException(type(e), e.pgcode, str(e))


